#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <stdlib.h>
#include <assert.h>
#include <richedit.h>
#include "defblib.h"

static WNDPROC s_wpOldDlg = 0;
static WNDPROC s_wpOldBtn = 0;

static BOOL isClass( HWND hwnd, LPCTSTR pszClass ) {
    TCHAR szClassName[ 100 ] = { 0 };
    GetClassName( 
        hwnd, szClassName, sizeof szClassName / sizeof( TCHAR ) );
    return 0 == _tcsicmp( szClassName, pszClass );
}

static BOOL isPushBtn( HWND hwndBtn ) {
    DWORD dwBtnStyle = 0xf & GetWindowLong( hwndBtn, GWL_STYLE );
    assert( isClass( hwndBtn, _T( "Button" ) ) );
    return 0 == (dwBtnStyle & ~(BS_DEFPUSHBUTTON | BS_PUSHBUTTON));
}

static void WINAPI setDfltLook( HWND hwndBtn, BOOL bDefault ) {
    assert( isClass( hwndBtn, _T( "Button" ) ) );
    assert( isPushBtn( hwndBtn ) );
    Button_SetStyle( hwndBtn, 
        bDefault ? BS_DEFPUSHBUTTON : BS_PUSHBUTTON, TRUE );
}

static BOOL isEditAndWantsReturn( HWND hwndEdt ) {
    if ( IsWindow( hwndEdt ) && isClass( hwndEdt, _T( "Edit" ) ) ) {
        const DWORD dwStyle = GetWindowLong( hwndEdt, GWL_STYLE );
        const BOOL wantsReturn = 0 != (ES_WANTRETURN & dwStyle);
        const BOOL isReadOnly  = 0 != (ES_READONLY   & dwStyle);
        assert( !(wantsReturn && isReadOnly) );
        return wantsReturn;
    }
    return FALSE;
}

static void enableDfltLook( HWND hwndDlg, BOOL bEnable ) {
    const UINT nID = CallWindowProc( 
        s_wpOldDlg, hwndDlg, DM_GETDEFID, 0, 0 );
    HWND hwndDef = DC_HASDEFID == HIWORD( nID ) ? 
        GetDlgItem( hwndDlg, LOWORD( nID ) ) : 0;
    if ( IsWindow( hwndDef ) ) {
        if ( !IsWindowEnabled( hwndDef ) ) {
            bEnable = FALSE;
        }
        setDfltLook( hwndDef, bEnable );
    }
}

static LRESULT CALLBACK 
myBtn( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {

    LRESULT lResult = 0;
    const BOOL canBeDefault = IsWindowEnabled( hwnd ) && 
        !isEditAndWantsReturn( GetFocus() );
    if ( !canBeDefault && BM_SETSTYLE == msg &&
         (BS_DEFPUSHBUTTON & wParam) ) 
    {
        wParam &= ~BS_DEFPUSHBUTTON;               // Disallowed!
    } 
    lResult = CallWindowProc( 
        s_wpOldBtn, hwnd, msg, wParam, lParam );
    if ( WM_ENABLE == msg && isPushBtn( hwnd ) ) {
        if ( (BOOL) wParam ) {                     // Enabling.
            HWND hwndDlg = GetParent( hwnd );
            if ( IsWindow( hwndDlg ) ) {
                const UINT nID = CallWindowProc( 
                    s_wpOldDlg, hwndDlg, DM_GETDEFID, 0, 0 );
                if ( DC_HASDEFID == HIWORD( nID ) ) {
                    const HWND hwndDef = GetDlgItem( 
                        hwndDlg, LOWORD( nID ) );
                    if ( hwnd == hwndDef ) {
                        setDfltLook( hwnd, TRUE );
                    }
                }
            }
        } else {
            setDfltLook( hwnd, FALSE );            // Disabling.
        }
    }
    return lResult;
}

static LRESULT CALLBACK 
myDlg( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam ) {

    int  codeNotify = GET_WM_COMMAND_CMD ( wParam, lParam );
    HWND hwndCtl    = GET_WM_COMMAND_HWND( wParam, lParam );
    
    if ( WM_COMMAND == msg ) {
        if ( CBN_DROPDOWN == codeNotify || 
             CBN_CLOSEUP  == codeNotify ) 
        {
            if ( isClass( hwndCtl, _T( "ComboBox" ) ) ) {
                enableDfltLook( hwnd, CBN_CLOSEUP == codeNotify );
            }
        } else if ( EN_SETFOCUS  == codeNotify || 
                    EN_KILLFOCUS == codeNotify ) 
        {
            if ( isEditAndWantsReturn( hwndCtl ) ) {
                enableDfltLook( hwnd, EN_KILLFOCUS == codeNotify );
            }
        }
    }
    return CallWindowProc( s_wpOldDlg, hwnd, msg, wParam, lParam );
}

static void __cdecl unFixDefBtns( void ) {
    if ( 0 != s_wpOldDlg ) {
        HWND hwnd = CreateWindow( WC_DIALOG, _T( "" ), WS_POPUP,
            0, 0, 0, 0, HWND_DESKTOP, 0, GetModuleHandle( 0 ), 0 );
        SetClassLong( hwnd, GCL_WNDPROC, (LONG) s_wpOldDlg );
        DestroyWindow( hwnd );
    }
    if ( 0 != s_wpOldBtn ) {
        HWND hwnd = CreateWindow( 
            _T( "Button" ), _T( "" ), WS_POPUP,
            0, 0, 0, 0, HWND_DESKTOP, 0, GetModuleHandle( 0 ), 0 );
        SetClassLong( hwnd, GCL_WNDPROC, (LONG) s_wpOldBtn );
        DestroyWindow( hwnd );
    }
}

void WINAPI fixDefBtns( void ) {
    HWND hwnd = CreateWindow( WC_DIALOG, _T( "" ), WS_POPUP, 
        0, 0, 0, 0, HWND_DESKTOP, 0, GetModuleHandle( 0 ), 0 );
    if ( IsWindow( hwnd ) ) {
        s_wpOldDlg = (WNDPROC) SetClassLong( 
            hwnd, GCL_WNDPROC, (LONG) myDlg );
        DestroyWindow( hwnd );
    }
    
    hwnd = CreateWindow( _T( "Button" ), _T( "" ), WS_POPUP, 
        0, 0, 0, 0, HWND_DESKTOP, 0, GetModuleHandle( 0 ), 0 );
    if ( IsWindow( hwnd ) ) {
        s_wpOldBtn = (WNDPROC) SetClassLong( 
            hwnd, GCL_WNDPROC, (LONG) myBtn );
        DestroyWindow( hwnd );
    }
    
    atexit( unFixDefBtns );
}
